
from vsg.token import process_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import process_declarative_part
from vsg.vhdlFile.classify import process_statement_part
from vsg.vhdlFile.classify import process_sensitivity_list


def detect(iToken, lObjects):
    '''
    process_statement ::=
        [ *process*_label : ]
            [ postponed ] process [ ( process_sensitivity_list ) ] [ is ]
                process_declarative_part
            begin
                process_statement_part
            end [ postponed ] process [ *process*_label ] ;
    '''
    if utils.find_in_next_n_tokens('process', 4, iToken, lObjects):
        if not utils.find_in_next_n_tokens(';', 3, iToken, lObjects):
            return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):
    iCurrent = classify_opening_declaration(iToken, lObjects)

    iCurrent = process_declarative_part.detect(iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('begin', token.begin_keyword, iCurrent, lObjects)

    iCurrent = process_statement_part.detect(iCurrent, lObjects)

    iCurrent = classify_closing_declaration(iCurrent, lObjects)

    return iCurrent


def classify_opening_declaration(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.process_label, token.label_colon)
    iCurrent = utils.assign_next_token_if('postponed', token.postponed_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('process', token.process_keyword, iCurrent, lObjects)

    if utils.is_next_token('(', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iCurrent, lObjects)
        iCurrent = process_sensitivity_list.classify(iCurrent, lObjects)
        iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)

    iCurrent = utils.assign_next_token_if('is', token.is_keyword, iCurrent, lObjects)

    return iCurrent


def classify_closing_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('postponed', token.end_postponed_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('process', token.end_process_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.end_process_label, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import blank_line_below_line_ending_with_token

from vsg import token

lTokens = []
lTokens.append(token.package_body.is_keyword)


class rule_201(blank_line_below_line_ending_with_token):
    '''
    This rule checks for a blank line below the **package** keyword.

    |configuring_blank_lines_link|

    **Violation**

    .. code-block:: vhdl

       package body FIFO_PKG is
         constant width : integer := 32;

    **Fix**

    .. code-block:: vhdl

       package body FIFO_PKG is

         constant width : integer := 32;
    '''

    def __init__(self):
        blank_line_below_line_ending_with_token.__init__(self, 'package_body', '201', lTokens)




from vsg.vhdlFile.extract import tokens


def get_tokens_bounded_by_unless_between(oStart, oEnd, lUnless, lAllObjects, oTokenMap):

    lReturn = []

    lStart, lEnd = oTokenMap.get_token_pair_indexes(oStart, oEnd)

    lUnlessStart, lUnlessEnd = get_unless_token_indexes(lUnless, oTokenMap)

    lNewStart, lNewEnd = filter_token_pairs(lStart, lEnd, lUnlessStart, lUnlessEnd)

    for iStart, iEnd, iIndex in zip(lNewStart, lNewEnd, lEnd):
        lTemp = lAllObjects[iStart: iEnd + 1]
        iStartLine = oTokenMap.get_line_number_of_index(iStart)
        oToi = tokens.New(iStart, iStartLine, lTemp)
        oToi.set_token_value(iIndex - iStart)
        lReturn.append(oToi)

    return lReturn


def get_unless_token_indexes(lUnless, oTokenMap):
    lReturnStart = []
    lReturnEnd = []
    for unless in lUnless:
        lStart, lEnd = oTokenMap.get_token_pair_indexes(unless[0], unless[1])
        lReturnStart.extend(lStart)
        lReturnEnd.extend(lEnd)
    return lReturnStart, lReturnEnd


def filter_token_pairs(lStart, lEnd, lUnlessStart, lUnlessEnd):
    lReturnStart = []
    lReturnEnd = []
    for iStart, iEnd in zip(lStart, lEnd):
        if not indexes_between_indexes(iStart, iEnd, lUnlessStart, lUnlessEnd):
            lReturnStart.append(iStart)
            lReturnEnd.append(iEnd)
    return lReturnStart, lReturnEnd


def indexes_between_indexes(iStart, iEnd, lUnlessStart, lUnlessEnd):
    for iUnlessStart, iUnlessEnd in zip(lUnlessStart, lUnlessEnd):
        if iUnlessStart < iStart < iUnlessEnd and \
           iUnlessStart < iEnd < iUnlessEnd:
            return True
    return False




from .rule_001 import rule_001
from .rule_002 import rule_002
from .rule_003 import rule_003
from .rule_004 import rule_004
from .rule_005 import rule_005
from .rule_006 import rule_006
from .rule_007 import rule_007
from .rule_008 import rule_008
from .rule_009 import rule_009
from .rule_010 import rule_010
from .rule_011 import rule_011
from .rule_012 import rule_012

from .rule_100 import rule_100
from .rule_101 import rule_101

from .rule_200 import rule_200
from .rule_201 import rule_201
from .rule_202 import rule_202
from .rule_203 import rule_203
from .rule_204 import rule_204
from .rule_205 import rule_205

from .rule_401 import rule_401
from .rule_410 import rule_410
from .rule_411 import rule_411
from .rule_412 import rule_412

from .rule_500 import rule_500
from .rule_501 import rule_501
from .rule_502 import rule_502
from .rule_503 import rule_503
from .rule_504 import rule_504
from .rule_505 import rule_505
from .rule_506 import rule_506
from .rule_507 import rule_507




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import loop_statement as token


class rule_103(Rule):
    '''
    This rule checks if a label exists that a single space exists between the label and the colon.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

         label: for index in 4 to 23 loop
         label    : for index in 0 to 100 loop

    **Fix**

    .. code-block:: vhdl

         label : for index in 4 to 23 loop
         label : for index in 0 to 100 loop
    '''
    def __init__(self):
        Rule.__init__(self, 'loop_statement', '103')
        self.left_token = token.loop_label
        self.right_token = token.label_colon





from vsg import parser
from vsg import token
from vsg import violation

from vsg.rule_group import alignment
from vsg.rules import alignment_utils
from vsg.rules import utils as rule_utils
from vsg.vhdlFile import utils


class align_tokens_in_region_between_tokens_unless_between_tokens(alignment.Rule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : token object list
       List of tokens to align

    left_token : token object
       The first token that defines the region

    right_token : token object
       The second token that defines the region

    lUnless : token object pairs list
       A list of pairs of tokens in which to exclude alignment
    '''

    def __init__(self, name, identifier, lTokens, left_token, right_token, lUnless):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens
        self.left_token = left_token
        self.right_token = right_token
        self.lUnless = lUnless
        ## attributes below are configurable by the user

        self.compact_alignment = True
        self.configuration.append('compact_alignment')

        self.blank_line_ends_group = True
        self.configuration.append('blank_line_ends_group')
        self.comment_line_ends_group = True
        self.configuration.append('comment_line_ends_group')

        self.if_control_statements_ends_group = False
        self.configuration.append('if_control_statements_ends_group')
        self.case_control_statements_ends_group = False
        self.configuration.append('case_control_statements_ends_group')
        self.loop_control_statements_ends_group = False
        self.configuration.append('loop_control_statements_ends_group')

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by_unless_between(self.left_token, self.right_token, self.lUnless)

    def analyze(self, oFile):
        lToi = self._get_tokens_of_interest(oFile)
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()
            iColumn = 0
            bTokenFound = False
            iToken = -1
            bSkip = False
            oEndSkipToken = None
            dAnalysis = {}

            for iIndex in range(0, len(lTokens)):
               iToken += 1
               oToken = lTokens[iIndex]

               bSkip, oEndSkipToken = check_for_exclusions(oToken, bSkip, oEndSkipToken, self.lUnless)

               if not bTokenFound and not bSkip:
                   for oSearch in self.lTokens:
                       if isinstance(oToken, oSearch):
                           bTokenFound = True
                           dAnalysis[iLine] = {}
                           dAnalysis[iLine]['token_column'] = iColumn
                           dAnalysis[iLine]['token_index'] = iToken
                           dAnalysis[iLine]['line_number'] = iLine
                           dAnalysis[iLine]['token_value'] = oToken.get_value()
                           if isinstance(lTokens[iIndex -1], parser.whitespace):
                               dAnalysis[iLine]['left_column'] = iColumn - len(lTokens[iIndex - 1].get_value())
                           else:
                               dAnalysis[iLine]['left_column'] = iColumn
                           break

                   iColumn += alignment_utils.update_column_width(self, oToken)

               if isinstance(oToken, parser.carriage_return):
                   iLine += 1
                   iColumn = 0
                   bTokenFound = False
                   iToken = -1
                   if self.comment_line_ends_group:
                       if utils.are_next_consecutive_token_types([parser.whitespace, parser.comment], iIndex + 1, lTokens) or \
                          utils.are_next_consecutive_token_types([parser.comment], iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}
                   if self.blank_line_ends_group:
                       if utils.are_next_consecutive_token_types([parser.blank_line], iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}

                   if self.if_control_statements_ends_group:
                       if check_for_if_keywords(iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}

                   if is_case_control_enabled(self.case_control_statements_ends_group):
                       if is_case_keyword(self.case_control_statements_ends_group, iIndex, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}

                   if self.loop_control_statements_ends_group:
                       if check_for_loop_keywords(iIndex + 1, lTokens):
                           add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)

                           for iKey in list(dAnalysis.keys()):
                               if dAnalysis[iKey]['adjust'] != 0:
                                   oLineTokens = oFile.get_tokens_from_line(iKey)
                                   sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                                   oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                                   oViolation.set_action(dAnalysis[iKey])
                                   self.add_violation(oViolation)

                           dAnalysis = {}



            add_adjustments_to_dAnalysis(dAnalysis, self.compact_alignment)


            for iKey in list(dAnalysis.keys()):
                if dAnalysis[iKey]['adjust'] != 0:
                    oLineTokens = oFile.get_tokens_from_line(iKey)
                    sSolution = 'Move ' + dAnalysis[iKey]['token_value'] + ' ' + str(dAnalysis[iKey]['adjust']) + ' columns'
                    oViolation = violation.New(oLineTokens.get_line_number(), oLineTokens, sSolution)
                    oViolation.set_action(dAnalysis[iKey])
                    self.add_violation(oViolation)

            dAnalysis = {}

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        iTokenIndex = dAction['token_index']

        if isinstance(lTokens[iTokenIndex - 1], parser.whitespace):
            iLen = len(lTokens[iTokenIndex - 1].get_value())
            lTokens[iTokenIndex - 1].set_value(' '*(iLen + dAction['adjust']))
        else:
            rule_utils.insert_whitespace(lTokens, iTokenIndex, dAction['adjust'])
        oViolation.set_tokens(lTokens)


def add_adjustments_to_dAnalysis(dAnalysis, compact_alignment):
    iMaxLeftColumn = 0
    iMinLeftColumn = 9999999999999999
    iMaxTokenColumn = 0
    iMinTokenColumn = 9999999999999999

    for iKey in list(dAnalysis.keys()):
        iMaxLeftColumn = max(iMaxLeftColumn, dAnalysis[iKey]['left_column'])
        iMinLeftColumn = min(iMinLeftColumn, dAnalysis[iKey]['left_column'])
        iMaxTokenColumn = max(iMaxTokenColumn, dAnalysis[iKey]['token_column'])
        iMinTokenColumn = min(iMinTokenColumn, dAnalysis[iKey]['token_column'])

    if compact_alignment:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxLeftColumn - dAnalysis[iKey]['token_column'] + 1
    else:
        for iKey in list(dAnalysis.keys()):
            dAnalysis[iKey]['adjust'] = iMaxTokenColumn - dAnalysis[iKey]['token_column']


def check_for_exclusions(oToken, bSkip, oEndSkipToken, lUnless):
    if bSkip:
        if isinstance(oToken, oEndSkipToken):
            return False, None
    else:
        for lTokenPairs in lUnless:
            if isinstance(oToken, lTokenPairs[0]):
                return True, lTokenPairs[1]

    return bSkip, oEndSkipToken


def check_for_if_keywords(iToken, lTokens):
    iMyToken = iToken
    if isinstance(lTokens[iToken], parser.whitespace):
        iMyToken += 1

    if isinstance(lTokens[iMyToken], token.if_statement.if_label):
        return True

    if isinstance(lTokens[iMyToken], token.if_statement.if_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.if_statement.elsif_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.if_statement.else_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.if_statement.end_keyword):
        return True

    return False

def check_for_case_keywords(iToken, lTokens):
    iMyToken = iToken
    if isinstance(lTokens[iToken], parser.whitespace):
        iMyToken += 1

    if isinstance(lTokens[iMyToken], token.case_statement.case_label):
        return True

    if isinstance(lTokens[iMyToken], token.case_statement.case_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.case_statement.end_keyword):
        return True

    return False

def check_for_when_keywords(iToken, lTokens):
    iMyToken = iToken
    if isinstance(lTokens[iToken], parser.whitespace):
        iMyToken += 1

    if isinstance(lTokens[iMyToken], token.case_statement_alternative.when_keyword):
        return True

    return False

def check_for_loop_keywords(iToken, lTokens):
    iMyToken = iToken
    if isinstance(lTokens[iToken], parser.whitespace):
        iMyToken += 1

    if isinstance(lTokens[iMyToken], token.loop_statement.loop_label):
        return True

    if isinstance(lTokens[iMyToken], token.iteration_scheme.while_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.iteration_scheme.for_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.loop_statement.loop_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.loop_statement.end_keyword):
        return True

    if isinstance(lTokens[iMyToken], token.loop_statement.end_loop_keyword):
        return True

    return False

def is_case_control_enabled(config):
    if config == 'break_on_case_or_end_case':
        return True
    return config

def is_case_keyword(config, iIndex, lTokens):
    if check_for_case_keywords(iIndex + 1, lTokens):
        return True
    if check_for_when_keywords(iIndex + 1, lTokens):
        if config != 'break_on_case_or_end_case':
           return True
    return False




from vsg import deprecated_rule


class rule_011(deprecated_rule.Rule):
    '''
    The function of this rule has been superceeded by the following rules:

    * ieee_500
    * subtype_002
    * type_014
    '''
    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'signal', '011')
        self.message.append('Rule ' + self.unique_id + ' has been superceeded by the following rules:')
        self.message.append('  ieee_500')
        self.message.append('  subtype_002')
        self.message.append('  type_014')




from vsg.token import type_mark as token

from vsg.vhdlFile import utils

from vsg import parser


def classify(iToken, lObjects):
    '''
    type_mark ::=
        *type*_name
      | *subtype*_name
    '''
    iCurrent = utils.assign_next_token(token.name, iToken, lObjects)

    if utils.is_next_token("'", iCurrent, lObjects):
        iCurrent = utils.assign_next_token(token.tic, iCurrent, lObjects)
        iCurrent = utils.assign_next_token(token.attribute, iCurrent, lObjects)

    return iCurrent




from vsg import parser


class shared_keyword(parser.keyword):
    '''
    unique_id = variable_declaration : shared_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class variable_keyword(parser.keyword):
    '''
    unique_id = variable_declaration : variable_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class keyword(parser.keyword):
    '''
    unique_id = variable_declaration : keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class identifier(parser.identifier):
    '''
    unique_id = variable_declaration : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class comma(parser.comma):
    '''
    unique_id = variable_declaration : comma
    '''

    def __init__(self):
        parser.comma.__init__(self)


class colon(parser.colon):
    '''
    unique_id = variable_declaration : colon
    '''

    def __init__(self, sString=':'):
        parser.colon.__init__(self)


class subtype_indication(parser.subtype_indication):
    '''
    unique_id = variable_declaration : subtype_indication
    '''

    def __init__(self, sString):
        parser.subtype_indication.__init__(self, sString)


class assignment_operator(parser.assignment):
    '''
    unique_id = variable_declaration : assignment_operator
    '''

    def __init__(self, sString=':='):
        parser.item.__init__(self, sString)


class assignment_expression(parser.item):
    '''
    unique_id = variable_declaration : assignment_expression
    '''

    def __init__(self, sString):
        parser.item.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = variable_declaration : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.case_generate_statement.label_colon, token.case_generate_statement.case_keyword])
lTokens.append([token.for_generate_statement.label_colon, token.for_generate_statement.for_keyword])
lTokens.append([token.if_generate_statement.label_colon, token.if_generate_statement.if_keyword])


class rule_014(Rule):
    '''
    This rule checks for a single space between the colon and the **for** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       ram_array :for i in 0 to 7 generate
       ram_array :   for i in 0 to 7 generate

    **Fix**

    .. code-block:: vhdl

       ram_array : for i in 0 to 7 generate
       ram_array : for i in 0 to 7 generate
    '''
    def __init__(self):
        Rule.__init__(self, 'generate', '014', lTokens)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens as Rule

from vsg import token

lTokens = []
lTokens.append(token.force_mode.out_keyword)
lTokens.append(token.force_mode.in_keyword)

lTokenPairs = []
lTokenPairs.append([token.selected_force_assignment.with_keyword, token.selected_force_assignment.semicolon])


class rule_007(Rule):
    '''
    This rule checks for code after the force mode keywords **in** and **out**.

    **Violation**

    .. code-block:: vhdl

       with mux_sel select addr <= force in "0000" when 0,
         "0001" when 1,
         "1111" when others;

    **Fix**

    .. code-block:: vhdl

       with mux_sel select addr <= force in
         "0000" when 0,
         "0001" when 1,
         "1111" when others;
    '''

    def __init__(self):
        Rule.__init__(self, 'selected_assignment', '007', lTokens, lTokenPairs)
        self.solution = 'Move code after the force mode keyword to the next line.'




from vsg.token import package_body as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import package_body_declarative_part


def detect(iToken, lObjects):
    '''
    package_body ::=
        package body *package*_simple_name is
            package_body_declarative_part
        end [ package body ] [ *package*_simple_name ] ;
    '''

    if utils.are_next_consecutive_tokens(['package', 'body', None, 'is'], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = classify_opening_declaration(iToken, lObjects)

    iCurrent = package_body_declarative_part.detect(iCurrent, lObjects)

    iCurrent = classify_closing_declaration(iToken, lObjects)

    return iCurrent


def classify_opening_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('package', token.package_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('body', token.body_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token(token.package_simple_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    return iCurrent


def classify_closing_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    if utils.are_next_consecutive_tokens(['package'], iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('package', token.end_package_keyword, iToken, lObjects)
        iCurrent = utils.assign_next_token_required('body', token.end_body_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.end_package_simple_name, iToken, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iToken, lObjects)

    return iCurrent




from .rule_001 import rule_001
from .rule_002 import rule_002

from .rule_100 import rule_100
from .rule_101 import rule_101

from .rule_300 import rule_300

from .rule_400 import rule_400

from .rule_500 import rule_500
from .rule_501 import rule_501




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.context_declaration.end_context_keyword)


class rule_015(token_case):
    '''
    This rule checks the context keyword has proper case in the end context declaration.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       end CONTEXT;

    **Fix**

    .. code-block:: vhdl

       end context;
    '''

    def __init__(self):
        token_case.__init__(self, 'context', '015', lTokens)
        self.groups.append('case::keyword')




from vsg.token import selected_expressions as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import choices
from vsg.vhdlFile.classify import expression


def classify_until(lUntils, iToken, lObjects):
    '''
    selected_expressions ::=
        { expression when choices , }
        expression when choices
    '''

    iCurrent = iToken
    lMyUntils = lUntils
    lMyUntils.append(',')

    iCurrent = expression.classify_until(['when'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('when', token.when_keyword, iCurrent, lObjects)

    iCurrent = choices.classify_until(lMyUntils, iCurrent, lObjects)

    while utils.is_next_token(',', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required(',', token.comma, iCurrent, lObjects)

        iCurrent = expression.classify_until(['when'], iCurrent, lObjects)

        iCurrent = utils.assign_next_token_required('when', token.when_keyword, iCurrent, lObjects)

        iCurrent = choices.classify_until(lMyUntils, iCurrent, lObjects)

    return iCurrent




from vsg.vhdlFile.classify import case_generate_statement
from vsg.vhdlFile.classify import for_generate_statement
from vsg.vhdlFile.classify import if_generate_statement


def detect(iToken, lObjects):
    '''
    generate_statement ::=
        for_generate_statement
      | if_generate_statement
      | case_generate_statement
    '''

    iCurrent = for_generate_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = if_generate_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = case_generate_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iCurrent




from vsg.rules import multiline_alignment_between_tokens as Rule

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.variable_declaration.variable_keyword, token.variable_declaration.semicolon])


class rule_400(Rule):
    '''
    This rule checks alignment of multiline constraints in variable declarations.

    |configuring_multiline_indent_rules_link|

    **Violation**

    .. code-block:: vhdl

       variable v_element : my_record(
                element1(7 downto 0),
       element2(3 downto 0)
               );

    **Fix**

    .. code-block:: vhdl

       variable v_element : my_record(
           element1(7 downto 0),
           element2(3 downto 0)
         );
    '''

    def __init__(self):
        Rule.__init__(self, 'variable', '400', lTokenPairs)
        self.phase = 5
        self.subphase = 3
        self.bIgnoreStartParen = True
        self.bConstraint = True





from vsg.rules import move_token_next_to_another_token as Rule

from vsg.token import full_type_declaration as token


class rule_018(Rule):
    '''
    This rule checks the **is** keyword is on the same line as the identifier.

    **Violation**

    .. code-block:: vhdl

       type t_record
       is

    **Fix**

    .. code-block:: vhdl

       type t_record is
    '''

    def __init__(self):
        Rule.__init__(self, 'type', '018', token.identifier, token.is_keyword)
        self.solution = 'Ensure *is* keyword is on the same line as the identifier.'
        self.subphase = 2




from vsg.token import exit_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import condition


def detect(iToken, lObjects):
    '''
    exit_statement ::=
        [ label : ] exit [ loop_label ] [ when condition ] ;
    '''

    if utils.are_next_consecutive_tokens([None, ':', 'exit'], iToken, lObjects):
        return classify(iToken, lObjects)
    if utils.is_next_token('exit', iToken, lObjects):
        return classify(iToken, lObjects)

    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.tokenize_label(iToken, lObjects, token.label, token.label_colon)

    iCurrent = utils.assign_next_token_required('exit', token.exit_keyword, iCurrent, lObjects)

    iCurrent = utils.assign_next_token_if_not_one_of([';', 'when'], token.loop_label, iCurrent, lObjects)

    if utils.is_next_token('when', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('when', token.when_keyword, iCurrent, lObjects)
        iCurrent = condition.classify_until([';'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg import parser
from vsg import token
from vsg import violation

from vsg.vhdlFile import utils
from vsg.rule_group import structure

lIfBoundingTokens = [token.if_statement.if_keyword, token.if_statement.then_keyword]

lElsifBoundingTokens = [token.if_statement.elsif_keyword, token.if_statement.then_keyword]

oStart = token.process_statement.begin_keyword
oEnd = token.process_statement.end_keyword


lAssignments = []
lAssignments.append(token.simple_waveform_assignment.assignment)
lAssignments.append(token.simple_force_assignment.assignment)
lAssignments.append(token.simple_release_assignment.assignment)

lEndAssignments = []
lEndAssignments.append(token.simple_waveform_assignment.semicolon)
lEndAssignments.append(token.simple_force_assignment.semicolon)
lEndAssignments.append(token.simple_release_assignment.semicolon)


class rule_001(structure.Rule):
    '''
    This rule checks for **after x** in signal assignments in clock processes.

    **Violation**

    .. code-block:: vhdl

       clk_proc : process(clock, reset) is
       begin
         if (reset = '1') then
           a <= '0';
           b <= '1';
         elsif (clock'event and clock = '1') then
           a <= d;
           b <= c;
         end if;
       end process clk_proc;

    **Fix**

    .. code-block:: vhdl

       clk_proc : process(clock, reset) is
       begin
         if (reset = '1') then
           a <= '0';
           b <= '1';
         elsif (clock'event and clock = '1') then
           a <= d after 1 ns;
           b <= c after 1 ns;
         end if;
       end process clk_proc;

    .. NOTE::  This rule has two configurable items:

       * magnitude
       * units

       The **magnitude** is the number of units.  Default is *1*.

       The **units** is a valid time unit: ms, us, ns, ps etc...  Default is *ns*.
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'after', '001')
        self.disable = True
        self.oStart = oStart
        self.oEnd = oEnd
        self.magnitude = 1
        self.units = 'ns'
        self.configuration.extend(['magnitude', 'units'])
        self.configuration_documentation_link = None

    def _get_tokens_of_interest(self, oFile):
        return oFile.get_tokens_bounded_by(self.oStart, self.oEnd)

    def _analyze(self, lToi):
        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)
            bInsideClockDef = False
            bInsideAssignment = False
            for iToken, oToken in enumerate(lTokens):
                iLine = utils.increment_line_number(iLine, oToken)

                if not bInsideClockDef:
                    bInsideClockDef = detect_clock_definition(iToken, oToken, lTokens)
                    continue

                if isinstance(oToken, token.if_statement.semicolon) and oToken.get_hierarchy() == 0:
                    bInsideClockDef = False
                    continue

                if not bInsideAssignment:
                    bInsideAssignment = detect_signal_assignment(oToken)
                    continue

                if bInsideAssignment:
                    if isinstance(oToken, token.waveform_element.after_keyword):
                        bInsideAssignment = False
                    if detect_end_signal_assignment(oToken):
                        oNewToi = oToi.extract_tokens(iToken, iToken)
                        sSolution = 'Add after ' + str(self.magnitude) + ' ' + self.units + ' to signal in clock process'
                        oViolation = violation.New(iLine, oNewToi, sSolution)
                        self.add_violation(oViolation)
                        bInsideAssignment = False

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        lNewTokens = []
        lNewTokens.append(parser.whitespace(' '))
        lNewTokens.append(token.waveform_element.after_keyword('after'))
        lNewTokens.append(parser.whitespace(' '))
        lNewTokens.append(parser.todo(str(self.magnitude)))
        lNewTokens.append(parser.whitespace(' '))
        lNewTokens.append(parser.todo(self.units))
        lNewTokens.extend(lTokens)

        oViolation.set_tokens(lNewTokens)


def detect_clock_definition(iToken, oToken, lTokens):
    if isinstance(oToken, token.if_statement.if_keyword) or isinstance(oToken, token.if_statement.elsif_keyword):
        if oToken.get_hierarchy() != 0:
            return False
        if utils.are_next_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis, token.ieee.std_logic_1164.function.rising_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([token.ieee.std_logic_1164.function.rising_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis, token.ieee.std_logic_1164.function.falling_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([token.ieee.std_logic_1164.function.falling_edge], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([None, parser.tic, token.predefined_attribute.event_keyword, token.logical_operator.and_operator, None, token.relational_operator.equal], iToken + 1, lTokens) or \
           utils.are_next_consecutive_token_types_ignoring_whitespace([parser.open_parenthesis, None, parser.tic, token.predefined_attribute.event_keyword, token.logical_operator.and_operator, None, token.relational_operator.equal], iToken + 1, lTokens):
            return True
    return False


def detect_signal_assignment(oToken):
    for oAssignment in lAssignments:
        if isinstance(oToken, oAssignment):
            return True
    return False


def detect_end_signal_assignment(oToken):
    for oEndAssignment in lEndAssignments:
        if isinstance(oToken, oEndAssignment):
            return True
    return False




from vsg.rules import token_case

from vsg.token import architecture_body as token


class rule_021(token_case):
    '''
    This rule checks the proper case of the **begin** keyword.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is
       BEGIN

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is
       begin
    '''

    def __init__(self):
        token_case.__init__(self, 'architecture', '021', [token.begin_keyword])
        self.groups.append('case::keyword')




from vsg import parser
from vsg import token

from vsg.rules import remove_lines_starting_with_token_between_token_pairs

lTokens = []
lTokens.append([token.simple_release_assignment.assignment, token.simple_release_assignment.semicolon])
lTokens.append([token.simple_force_assignment.assignment, token.simple_force_assignment.semicolon])
lTokens.append([token.simple_waveform_assignment.assignment, token.simple_waveform_assignment.semicolon])

oRemoveToken = parser.comment


class rule_006(remove_lines_starting_with_token_between_token_pairs):
    '''
    This rule checks for comments within multiline sequential statements.

    **Violation**

    .. code-block:: vhdl

       overflow <= wr_en and
        --         rd_address(0)
                   rd_en;

    **Fix**

    .. code-block:: vhdl

       overflow <= wr_en and
                   rd_en;
    '''

    def __init__(self):
        remove_lines_starting_with_token_between_token_pairs.__init__(self, 'sequential', '006', oRemoveToken, lTokens)
        self.solution = 'Remove comments inside sequential assignment'
        self.fixable = False




from vsg import token

from vsg.rules import remove_carriage_returns_between_token_pairs

lTokens = []
lTokens.append([token.for_generate_statement.generate_label, token.for_generate_statement.for_keyword])
lTokens.append([token.if_generate_statement.generate_label, token.if_generate_statement.if_keyword])
lTokens.append([token.case_generate_statement.generate_label, token.case_generate_statement.case_keyword])


class rule_015(remove_carriage_returns_between_token_pairs):
    '''
    This rule checks the generate label and the **generate** keyword are on the same line.
    Keeping the label and generate on the same line reduces excessive indenting.

    **Violation**

    .. code-block:: vhdl

       ram_array :
         for i in 0 to 7 generate

    **Fix**

    .. code-block:: vhdl

       ram_array : for i in 0 to 7 generate
    '''

    def __init__(self):
        remove_carriage_returns_between_token_pairs.__init__(self, 'generate', '015', lTokens)
        self.solution = 'Merge lines with generate label with *generate* keyword.'




from vsg.rules import multiline_alignment_between_tokens as Rule

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.procedure_call.procedure_name, token.procedure_call.close_parenthesis])


class rule_400(Rule):
    '''
    This rule checks the alignment of multiline procedure calls.

    |configuring_multiline_indent_rules_link|

    **Violation**

    .. code-block:: vhdl

       connect_ports(port_1, port_2, port_3,
             port_4, port_5,
                          port_6, port_7);

    **Fix**

    .. code-block:: vhdl

       connect_ports(port_1, port_2, port_3,
                     port_4, port_5,
                     port_6, port_7);
    '''

    def __init__(self):
        Rule.__init__(self, 'procedure_call', '400', lTokenPairs)
        self.phase = 5




from vsg.vhdlFile.extract import tokens


def get_tokens_from_beginning_of_line_containing_token_to_the_next_non_whitespace_token_to_the_right(token, lAllTokens, oTokenMap):

    lReturn = []

    lToken = oTokenMap.get_token_indexes(token)

    for iToken in lToken:
        iLine = oTokenMap.get_line_number_of_index(iToken)
        iStart = oTokenMap.get_index_of_carriage_return_before_index(iToken)
        iEnd = oTokenMap.get_index_of_next_non_whitespace_token_after_index_ignoring_comments(iToken)

        oTokens = tokens.New(iStart, iLine, lAllTokens[iStart:iEnd + 1])
        oTokens.set_meta_data('iTokenIndex', iToken - iStart)
        lReturn.append(oTokens)

    return lReturn
